package net.sourceforge.jenesis4java.impl;

/**
 * Copyright (C) 2008, 2010 Richard van Nieuwenhoven - ritchie [at] gmx [dot] at
 * Copyright (C) 2000, 2001 Paul Cody Johnston - pcj@inxar.org <br>
 * This file is part of Jenesis4java. Jenesis4java is free software: you can
 * redistribute it and/or modify it under the terms of the GNU Lesser General
 * Public License as published by the Free Software Foundation, either version 3
 * of the License, or (at your option) any later version.<br>
 * Jenesis4java is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
 * details.<br>
 * You should have received a copy of the GNU Lesser General Public License
 * along with Jenesis4java. If not, see <http://www.gnu.org/licenses/>.
 */
import java.util.ArrayList;
import java.util.List;

import net.sourceforge.jenesis4java.Assign;
import net.sourceforge.jenesis4java.Block;
import net.sourceforge.jenesis4java.Break;
import net.sourceforge.jenesis4java.Case;
import net.sourceforge.jenesis4java.Catch;
import net.sourceforge.jenesis4java.CodeWriter;
import net.sourceforge.jenesis4java.Codeable;
import net.sourceforge.jenesis4java.Comment;
import net.sourceforge.jenesis4java.ConditionalStatement;
import net.sourceforge.jenesis4java.Continue;
import net.sourceforge.jenesis4java.Default;
import net.sourceforge.jenesis4java.DoWhile;
import net.sourceforge.jenesis4java.Else;
import net.sourceforge.jenesis4java.ElseIf;
import net.sourceforge.jenesis4java.Empty;
import net.sourceforge.jenesis4java.Expression;
import net.sourceforge.jenesis4java.ExpressionStatement;
import net.sourceforge.jenesis4java.Finally;
import net.sourceforge.jenesis4java.For;
import net.sourceforge.jenesis4java.FormalParameter;
import net.sourceforge.jenesis4java.If;
import net.sourceforge.jenesis4java.Let;
import net.sourceforge.jenesis4java.LocalBlock;
import net.sourceforge.jenesis4java.LocalClass;
import net.sourceforge.jenesis4java.Return;
import net.sourceforge.jenesis4java.Statement;
import net.sourceforge.jenesis4java.Switch;
import net.sourceforge.jenesis4java.Synchronized;
import net.sourceforge.jenesis4java.Throw;
import net.sourceforge.jenesis4java.Try;
import net.sourceforge.jenesis4java.Type;
import net.sourceforge.jenesis4java.While;
import net.sourceforge.jenesis4java.impl.util.BlockStyle;
import net.sourceforge.jenesis4java.impl.util.ListTypeSelector;

/**
 * Standard <code>Statement</code> implementations.
 */
public abstract class MStatement extends MVM.MCodeable implements Statement {

    // ===============================================================
    // BLOCK STATEMENT
    // ===============================================================
    static class BlockStatement extends MStatement implements Block {

        List<Statement> vs;

        BlockStatement(MVM vm) {
            super(vm);
            this.vs = new ArrayList<Statement>();
        }

        public List<Statement> getStatements() {
            return ListTypeSelector.select(this.vs);
        }

        public Break newBreak() {
            Break x = new MBreak(this.vm);
            this.vs.add(x);
            return x;
        }

        public Continue newContinue() {
            Continue x = new MContinue(this.vm);
            this.vs.add(x);
            return x;
        }

        public Let newDeclarationLet(Type type) {
            Let x = new MStatement.MLet(this.vm, type);
            int index = 0;
            for (; index < this.vs.size() && this.vs.get(index) instanceof Let; index++) {
                ;
            }
            this.vs.add(index, x);
            return x;
        }

        public DoWhile newDoWhile(Expression predicate) {
            DoWhile x = new MDoWhile(this.vm, predicate);
            this.vs.add(x);
            return x;
        }

        public Empty newEmpty() {
            Empty x = new MEmpty(this.vm);
            this.vs.add(x);
            return x;
        }

        public For newFor() {
            For x = new MFor(this.vm);
            this.vs.add(x);
            return x;
        }

        public If newIf(Expression predicate) {
            If x = new MIf(this.vm, predicate);
            this.vs.add(x);
            return x;
        }

        public Let newLet(Type type) {
            Let x = new MLet(this.vm, type);
            this.vs.add(x);
            return x;
        }

        public LocalBlock newLocalBlock() {
            LocalBlock x = new MLocalBlock(this.vm);
            this.vs.add(x);
            return x;
        }

        public LocalClass newLocalClass(String name) {
            LocalClass x = new MDeclaration.MLocalClass(this.vm, name);
            this.vs.add(x);
            return x;
        }

        public Return newReturn() {
            Return x = new MReturn(this.vm);
            this.vs.add(x);
            return x;
        }

        public ExpressionStatement newStmt(Expression expr) {
            ExpressionStatement x = new MExpressionStatement(this.vm, expr);
            this.vs.add(x);
            return x;
        }

        public Switch newSwitch(Expression integer) {
            Switch x = new MSwitch(this.vm, integer);
            this.vs.add(x);
            return x;
        }

        public Synchronized newSynchronized(Expression mutex) {
            Synchronized x = new MSynchronized(this.vm, mutex);
            this.vs.add(x);
            return x;
        }

        public Throw newThrow(Expression throwable) {
            Throw x = new MThrow(this.vm, throwable);
            this.vs.add(x);
            return x;
        }

        public Try newTry() {
            Try x = new MTry(this.vm);
            this.vs.add(x);
            return x;
        }

        public While newWhile(Expression predicate) {
            While x = new MWhile(this.vm, predicate);
            this.vs.add(x);
            return x;
        }

        protected void writeBlock(CodeWriter out, BlockStyle style) {
            // write it
            style.toCode(out, getStatements());
        }
    }

    // ===============================================================
    // GOTO STATMENTS
    // ===============================================================
    abstract static class GotoStatement extends MStatement {

        String kwd;

        String target;

        GotoStatement(MVM vm, String kwd) {
            super(vm);
            this.kwd = kwd;
        }

        public String getGoto() {
            return this.target;
        }

        public GotoStatement setGoto(String target) {
            this.target = target;
            return this;
        }

        @Override
        public CodeWriter toCode(CodeWriter out) {
            super.toCode(out);

            out.write(this.kwd);

            if (this.target != null) {
                out.space().write(this.target);
            }

            return out.write(';');
        }
    }

    // ===============================================================
    // BREAK STATEMENT
    // ===============================================================
    static class MBreak extends GotoStatement implements Break {

        MBreak(MVM vm) {
            super(vm, "break");
        }

        @Override
        public MBreak setGoto(String target) {
            super.setGoto(target);
            return this;
        }
    }

    // ===============================================================
    // CASE STATEMENT
    // ===============================================================
    static class MCase extends MConditionalStatement implements Case {

        MCase(MVM vm, Expression expr) {
            super(vm);
            setPredicate(expr);
        }

        public Expression getConstant() {
            return getPredicate();
        }

        public MCase setConstant(Expression expr) {
            setPredicate(expr);
            return this;
        }

        @Override
        public CodeWriter toCode(CodeWriter out) {
            super.toCode(out);
            out.write("case ").write(this.e).write(':');
            writeBlock(out, this.vm.getStyle("case"));
            return out;
        }
    }

    // ===============================================================
    // CATCH STATEMENT
    // ===============================================================
    static class MCatch extends BlockStatement implements Catch {

        FormalParameter fp;

        MCatch(MVM vm, Type type, String name) {
            super(vm);
            this.fp = new MDeclaration.MFormalParameter(vm, type, name);
        }

        public FormalParameter getThrowable() {
            return this.fp;
        }

        public MCatch setThrowable(Type type, String name) {
            this.fp = new MDeclaration.MFormalParameter(this.vm, type, name);
            return this;
        }

        @Override
        public CodeWriter toCode(CodeWriter out) {
            super.toCode(out);
            out.write("catch (").write(this.fp).write(')');
            writeBlock(out, this.vm.getStyle("catch"));
            return out;
        }
    }

    // ===============================================================
    // CONDITIONAL STATEMENT
    // ===============================================================
    abstract static class MConditionalStatement extends BlockStatement implements ConditionalStatement {

        Expression e;

        public MConditionalStatement(MVM vm) {
            super(vm);
        }

        public MConditionalStatement(MVM vm, Expression e) {
            super(vm);
            this.e = e;
        }

        public Expression getPredicate() {
            return this.e;
        }

        public MConditionalStatement setPredicate(Expression e) {
            this.e = e;
            return this;
        }
    }

    // ===============================================================
    // CONTINUE STATEMENT
    // ===============================================================
    static class MContinue extends GotoStatement implements Continue {

        MContinue(MVM vm) {
            super(vm, "continue");
        }

        @Override
        public MContinue setGoto(String target) {
            super.setGoto(target);
            return this;
        }
    }

    // ===============================================================
    // DEFAULT STATEMENT
    // ===============================================================
    static class MDefault extends BlockStatement implements Default {

        MDefault(MVM vm) {
            super(vm);
        }

        @Override
        public CodeWriter toCode(CodeWriter out) {
            super.toCode(out);
            out.write("default:");
            writeBlock(out, this.vm.getStyle("default"));
            return out;
        }
    }

    // ===============================================================
    // DO WHILE STATEMENT
    // ===============================================================
    static class MDoWhile extends MConditionalStatement implements DoWhile {

        public MDoWhile(MVM vm, Expression e) {
            super(vm, e);
        }

        @Override
        public CodeWriter toCode(CodeWriter out) {
            super.toCode(out);
            out.write("do");
            writeBlock(out, this.vm.getStyle("do"));
            out.write("while (").write(this.e).write(");");
            return out;
        }
    }

    // ===============================================================
    // ELSE STATEMENT
    // ===============================================================
    static class MElse extends BlockStatement implements Else {

        MElse(MVM vm) {
            super(vm);
        }

        @Override
        public CodeWriter toCode(CodeWriter out) {
            super.toCode(out);
            out.write("else");
            writeBlock(out, this.vm.getStyle("else"));
            return out;
        }
    }

    // ===============================================================
    // ELSE IF STATEMENT
    // ===============================================================
    static class MElseIf extends MConditionalStatement implements ElseIf {

        MElseIf(MVM vm, Expression e) {
            super(vm, e);
        }

        @Override
        public CodeWriter toCode(CodeWriter out) {
            super.toCode(out);
            out.write("else if (").write(this.e).write(')');
            writeBlock(out, this.vm.getStyle("else-if"));
            return out;
        }
    }

    // ===============================================================
    // EMPTY STATEMENT
    // ===============================================================
    static class MEmpty extends MStatement implements Empty {

        MEmpty(MVM vm) {
            super(vm);
        }

        @Override
        public CodeWriter toCode(CodeWriter out) {
            super.toCode(out);
            out.write(';');
            return out;
        }
    }

    // ===============================================================
    // EXPRESSIONABLE STATEMENT
    // ===============================================================
    abstract static class MExpressionableStatement extends MStatement {

        Expression e;

        MExpressionableStatement(MVM vm) {
            super(vm);
        }

        public Expression getExpression() {
            return this.e;
        }

        public MExpressionableStatement setExpression(Expression e) {
            this.e = e;
            return this;
        }
    }

    // ===============================================================
    // EXPRESSION STATEMENT
    // ===============================================================
    static class MExpressionStatement extends MExpressionableStatement implements ExpressionStatement {

        Expression e;

        MExpressionStatement(MVM vm) {
            super(vm);
        }

        MExpressionStatement(MVM vm, Expression e) {
            super(vm);
            this.e = e;
        }

        @Override
        public Expression getExpression() {
            return this.e;
        }

        @Override
        public MExpressionStatement setExpression(Expression e) {
            this.e = e;
            return this;
        }

        @Override
        public CodeWriter toCode(CodeWriter out) {
            super.toCode(out);

            out.write(this.e).write(';');

            return out;
        }
    }

    // ===============================================================
    // CATCH STATEMENT
    // ===============================================================
    static class MFinally extends BlockStatement implements Finally {

        MFinally(MVM vm) {
            super(vm);
        }

        @Override
        public CodeWriter toCode(CodeWriter out) {
            super.toCode(out);
            out.write("finally");
            writeBlock(out, this.vm.getStyle("finally"));
            return out;
        }
    }

    // ===============================================================
    // FOR STATEMENT
    // ===============================================================
    static class MFor extends MConditionalStatement implements For {

        List<Codeable> vi;

        List<Codeable> vu;

        public MFor(MVM vm) {
            super(vm);
            this.vi = new ArrayList<Codeable>();
            this.vu = new ArrayList<Codeable>();
        }

        public MFor addInit(Expression e) {
            this.vi.add(e);
            return this;
        }

        public MFor addUpdate(Expression e) {
            this.vu.add(e);
            return this;
        }

        public List<Codeable> getInits() {
            return ListTypeSelector.select(this.vi);
        }

        public List<Codeable> getUpdates() {
            return ListTypeSelector.select(this.vu);
        }

        public Let setInit(Type type) {
            Let let = new MForInitLet(this.vm, type);
            this.vi.add(let);
            return let;
        }

        @Override
        public CodeWriter toCode(CodeWriter out) {
            super.toCode(out);
            out.write("for (");
            write(this.vi, out);
            out.write(';');
            if (this.e != null) {
                out.space().write(this.e);
            }
            out.write(';');
            if (this.vu.size() > 0) {
                out.space();
                write(this.vu, out);
            }
            out.write(')');
            writeBlock(out, this.vm.getStyle("for"));
            return out;
        }

        private void write(List<Codeable> v, CodeWriter out) {
            for (int i = 0; i < v.size(); i++) {
                if (i > 0) {
                    out.write(", ");
                }
                out.write(v.get(i));
            }
        }

    }

    // ===============================================================
    // LET STATEMENT
    // ===============================================================
    static class MForInitLet extends MLet {

        MForInitLet(MVM vm, Type type) {
            super(vm, type);
        }

        @Override
        public CodeWriter toCode(CodeWriter out) {
            // queue up the comment
            out.queue(this.comment);

            if (this.isFinal) {
                out.write("final").space();
            }

            out.write(this.type).space();

            for (int i = 0; i < this.v.size(); i++) {
                if (i > 0) {
                    out.write(',').space();
                }
                out.write(this.v.get(i));
            }

            return out;
        }
    }

    // ===============================================================
    // IF STATEMENT
    // ===============================================================
    static class MIf extends MConditionalStatement implements If {

        List<ElseIf> vcei;

        Else _else;

        MIf(MVM vm, Expression e) {
            super(vm, e);
            this.vcei = new ArrayList<ElseIf>();
        }

        public Else getElse() {
            if (this._else == null) {
                this._else = new MElse(this.vm);
            }
            return this._else;
        }

        public List<ElseIf> getElseIfs() {
            return ListTypeSelector.select(this.vcei);
        }

        public ElseIf newElseIf(Expression e) {
            ElseIf ei = new MElseIf(this.vm, e);
            this.vcei.add(ei);
            return ei;
        }

        @Override
        public CodeWriter toCode(CodeWriter out) {
            super.toCode(out);
            out.write("if (").write(this.e).write(')');
            writeBlock(out, this.vm.getStyle("if"));
            codeElseIfs(this.vcei, out);
            if (this._else != null && this._else.getStatements().size() > 0) {
                out.write(this._else);
            }
            return out;
        }

        private CodeWriter codeElseIfs(List<ElseIf> v, CodeWriter out) {
            for (ElseIf elseIf : v) {
                out.write(elseIf);
            }
            return out;
        }
    }

    // ===============================================================
    // LET STATEMENT
    // ===============================================================
    static class MLet extends MExpressionableStatement implements Let {

        Type type;

        boolean isFinal;

        List<Assign> v;

        MLet(MVM vm, Type type) {
            super(vm);
            this.type = type;
            this.v = new ArrayList<Assign>();
        }

        public MLet addAssign(Assign assign) {
            this.v.add(assign);
            return this;
        }

        public MLet addAssign(String name, Expression expr) {
            this.v.add(new MExpression.MAssign(this.vm, Assign.S, new MExpression.MVariable(this.vm, name), expr));
            return this;
        }

        public List<Assign> getAssigns() {
            return ListTypeSelector.select(this.v);
        }

        public Type getType() {
            return this.type;
        }

        public boolean isFinal() {
            return this.isFinal;
        }

        public MLet isFinal(boolean value) {
            this.isFinal = value;
            return this;
        }

        public MLet setType(Type type) {
            this.type = type;
            return this;
        }

        @Override
        public CodeWriter toCode(CodeWriter out) {
            super.toCode(out);

            if (this.isFinal) {
                out.write("final").space();
            }

            out.write(this.type).space();

            for (int i = 0; i < this.v.size(); i++) {
                if (i > 0) {
                    out.write(',').space();
                }
                out.write(this.v.get(i));
            }

            out.write(';');

            return out;
        }
    }

    // ===============================================================
    // LOCAL BLOCK STATEMENT
    // ===============================================================
    static class MLocalBlock extends BlockStatement implements LocalBlock {

        public MLocalBlock(MVM vm) {
            super(vm);
        }

        @Override
        public CodeWriter toCode(CodeWriter out) {
            super.toCode(out);
            writeBlock(out, this.vm.getStyle("local-class"));
            return out;
        }

    }

    // ===============================================================
    // RETURN STATEMENT
    // ===============================================================
    static class MReturn extends MExpressionableStatement implements Return {

        MReturn(MVM vm) {
            super(vm);
        }

        @Override
        public MReturn setExpression(Expression e) {
            super.setExpression(e);
            return this;
        }

        @Override
        public CodeWriter toCode(CodeWriter out) {
            super.toCode(out);

            out.write("return");

            if (this.e != null) {
                out.space().write(this.e);
            }

            out.write(';');

            return out;
        }
    }

    // ===============================================================
    // SWITCH STATEMENT
    // ===============================================================
    static class MSwitch extends MConditionalStatement implements Switch {

        boolean triggered;

        Default d;

        MSwitch(MVM vm, Expression e) {
            super(vm, e);
        }

        public List<Case> getCases() {
            return ListTypeSelector.select(this.vs, Case.class);
        }

        public Default getDefault() {
            if (this.d == null) {
                this.d = new MDefault(this.vm);
            }
            return this.d;
        }

        public Case newCase(Expression constant) {
            Case c = new MCase(this.vm, constant);
            this.vs.add(c);
            return c;
        }

        @Override
        public CodeWriter toCode(CodeWriter out) {
            if (!this.triggered) {
                // add in the default if its not null
                if (this.d != null) {
                    this.vs.add(this.d);
                }
                this.triggered = true;
            }

            super.toCode(out);
            out.write("switch (").write(this.e).write(')');
            writeBlock(out, this.vm.getStyle("switch"));
            return out;
        }
    }

    // ===============================================================
    // SYNCHRONIZED STATEMENT
    // ===============================================================
    static class MSynchronized extends BlockStatement implements Synchronized {

        Expression mutex;

        MSynchronized(MVM vm, Expression mutex) {
            super(vm);
            this.mutex = mutex;
        }

        public Expression getMutex() {
            return this.mutex;
        }

        public MSynchronized setMutex(Expression mutex) {
            this.mutex = mutex;
            return this;
        }

        @Override
        public CodeWriter toCode(CodeWriter out) {
            super.toCode(out);
            out.write("synchronized ").write('(').write(this.mutex).write(')');
            writeBlock(out, this.vm.getStyle("synchronized"));
            return out;
        }
    }

    // ===============================================================
    // THROW STATEMENT
    // ===============================================================
    static class MThrow extends MStatement implements Throw {

        Expression e;

        MThrow(MVM vm, Expression e) {
            super(vm);
            this.e = e;
        }

        public Expression getThrowable() {
            return this.e;
        }

        public MThrow setThrowable(Expression e) {
            this.e = e;
            return this;
        }

        @Override
        public CodeWriter toCode(CodeWriter out) {
            super.toCode(out);

            out.write("throw ").write(this.e).write(';');

            return out;
        }
    }

    // ===============================================================
    // TRY STATEMENT
    // ===============================================================
    static class MTry extends BlockStatement implements Try {

        List<Catch> vcc;

        Finally finallyClause;

        MTry(MVM vm) {
            super(vm);
            this.vcc = new ArrayList<Catch>();
        }

        public List<Catch> getCatches() {
            return ListTypeSelector.select(this.vcc);
        }

        public Finally getFinally() {
            if (this.finallyClause == null) {
                this.finallyClause = new MFinally(this.vm);
            }
            return this.finallyClause;
        }

        public Catch newCatch(Type type, String name) {
            Catch cc = new MCatch(this.vm, type, name);
            this.vcc.add(cc);
            return cc;
        }

        @Override
        public CodeWriter toCode(CodeWriter out) {
            super.toCode(out);
            out.write("try");
            writeBlock(out, this.vm.getStyle("try"));
            writeCatches(this.vcc, out);
            out.write(this.finallyClause);
            return out;
        }

        // utility method
        private CodeWriter writeCatches(List<Catch> v, CodeWriter out) {
            for (Catch catch1 : v) {
                out.write(catch1);
            }
            return out;
        }
    }

    // ===============================================================
    // WHILE STATEMENT
    // ===============================================================
    static class MWhile extends MConditionalStatement implements While {

        public MWhile(MVM vm, Expression e) {
            super(vm, e);
        }

        @Override
        public CodeWriter toCode(CodeWriter out) {
            super.toCode(out);
            out.write("while (").write(this.e).write(')');
            writeBlock(out, this.vm.getStyle("while"));
            return out;
        }
    }

    String label;

    public MStatement(MVM vm) {
        super(vm);
    }

    public Comment comment(String text) {
        Comment sl = new MComment.MSingleLineComment(this.vm, text);
        this.comment = sl;
        return sl;
    }

    public String getLabel() {
        return this.label;
    }

    public boolean isAbruptCompletionStatement() {
        return this instanceof Return || //
                this instanceof Break || //
                this instanceof Throw || //
                this instanceof Continue;
    }

    public MStatement setLabel(String label) {
        this.label = label;
        return this;
    }

    @Override
    public CodeWriter toCode(CodeWriter out) {
        // write the comment
        out.write(this.comment);

        // make a new line if this one is not new
        if (!out.isLineNew()) {
            out.newLine();
        }

        // and the label, if there is one.
        if (this.label != null) {
            out.write(this.label).write(':').space();
        }

        return out;
    }
}
